! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: emlmerge_mod
!
! Authors:  Brendan Perry - NASA/GSFC (rewrite) brendan.h.perry@nasa.gov
!           Steve Snowden - NASA/GSFC (original algorithm)
!
! Description: Extended Source Analysis Software
!
! Comments:
!
! $Id: eml_mod.f90,v 1.4 2009/05/26 12:53:39 aibarra Exp $
!
!  Merges emllist.fits source lists from emldetect
!
!  EMLMERGE creates a master list of sources detected from a list
!  of observations which have been processed to the point of a creating
!  an emldetect source list.  Sources are compared and those within 2"
!  of each other have the one with the lower flux thrown out.  It assumes
!  that the source detection has been in two bands, 0.4-1.25 keV and 
!  2.0-7.2 keV, which can be produced by cheese-bands in the default mode.
!  This task was originally called merge_source_list
!
!  INPUT:
!    dirfile     : Filename of directory list
!    maxlikelim  : Lower ML limit for source detections
!    runemlfill  : Run emlfill to fix NaN fluxes?
!
!    Example: emlmerge dirfile=dir.txt maxlikelim=15.0 runemlfill=T
!
!  OUTPUT:
!    Merged emllist_merged.fits source list

module emlmerge_mod

  use dal
  use param 
  use errorHandling
  use gofcoords_mod ! from goflib, formerly in subsn (SS)
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  use gofutils_mod  ! from goflib, formerly in subsn (SS)
  use, intrinsic :: ieee_arithmetic

  implicit none

  private

  character(len=*), parameter :: &
       dirfilepar = "dirfile", &     ! text directory list
       mlimpar    = "maxlikelim", &  ! lower ML limit for source detections
       runfillpar = "runemlfill"     ! run emllist if not done already?

  public emlmerge

contains

  !---------------------------------------------------------------------
  subroutine emlmerge
    !---------------------------------------------------------------------

    type(DataSetT) :: inEML_S, outEML_S
    type(TableT)   :: inEML_T, outEML_T
    type(ColumnT)  :: mlidsrc_C, idinst_C, idband_C, scts_C, &
         detml_C, flux_C, fluxerr_C, rate_C, rateerr_C, ra_C, dec_C, &
         radec_C, dist_C, outRA_C, outDEC_C, outRADEC_C, outML_C, &
         outFLUX_C, outFLUXS_C, outFLUXH_C, outFLUXERR_C, outFLUXSERR_C, &
         outFLUXHERR_C, outDIST_C
    !
    character(len=256) :: dirs_F, outEML_F, outQDP_F, currpath_F, currpath_FF
    character(len=256), dimension(:), allocatable :: fullpath, goodemllist
    character(len=256) :: currdir, inEML_F
    character(len=20) :: command, eml_F, eml_FF
    integer :: i, ii, j, k, ios, lunin, lunout, nRows, outRows, err, status
    integer :: crow, csrc, nGood, init
    integer, dimension(:), allocatable :: hist1, hist2
    integer, dimension(:), pointer :: mlidsrc_JP, idinst_JP, idband_JP
    logical :: runemlfill, warndie, exists, doit
    real(kind=single) :: maxlim, punc
    real(kind=single), dimension (:), allocatable :: dist, distt, &
         flux, fluxerr, flux1, flux1err, flux2, flux2err, fluxt, &
         fluxterr, rate, rateerr, rate1, rate1err, rate2, rate2err, &
         ml, mlt, ratet, rateterr, radec, raddect, raddecterr, &
         scts, sctst
    real(kind=single), dimension(:), allocatable :: outRADEC, outML, &
         outFLUX, outFLUXERR, outFLUXS, outFLUXSERR, outFLUXH, outFLUXHERR, &
         outDIST
    real(kind=single) :: tnf1, tnf2, tnr1, tnr2, tf1, tfe1, tf2, tfe2, &
         tr1, tre1, tr2, tre2
    real(kind=single) :: decave, decmax, decmin, delra, deldec,  histmax, &
         histrange, maxlikelim, raave, ramax, ramin, rarange, scale
    real(kind=single), dimension(:), pointer :: scts_RP, detml_RP, flux_RP, &
         fluxerr_RP, rate_RP, rateerr_RP, radec_RP, dist_RP, &
         outRADEC_RP, outML_RP, outFLUX_RP, outFLUXERR_RP, outFLUXS_RP, &
         outFLUXSERR_RP, outFLUXH_RP, outFLUXHERR_RP, outDIST_RP
    real(kind=double) :: dang, dlat1, dlat2, dlong1, dlong2
    real(kind=double), dimension(:), allocatable :: outRA, outDEC
    real(kind=double), dimension(:), allocatable :: ra, dec, rat, dect
    real(kind=double), dimension(:), pointer :: ra_DP, dec_DP, outRA_DP, outDEC_DP
    !
    ! get the parameters from the par file
    !
    dirs_F     = stringParameter(dirfilepar)
    runemlfill = booleanParameter(runfillpar)
    maxlim     = realParameter(mlimpar)
    !
    ! params not mandatory. if not given, it defaults to these generic names.
    !
    eml_F     = "emllist.fits"
    eml_FF    = "emllist_filled.fits"
    outEML_F  = "emllist_merged.fits"
    if (runemlfill) then ! run sas task emlfill 
       outEML_F    = 'emllist_filled_merged.fits'
    endif
    ! check final outfile for existence, clobber verdict
    warndie = .TRUE.
    doit = .FALSE.
    call inquirer(outEML_F,warndie,doit) ! will die gracefully if can't
    !
    write(errstr,*) "Directory list text file   = ",trim(adjustl(dirs_F))
    call message(AppMsg,Always)
    write(errstr,*) "Output merged source list  = ",trim(adjustl(outEML_F))
    call message(AppMsg,Always)
    write(errstr,*) "Automatically run emlfill? = ",runemlfill
    call message(AppMsg,Always)
    write(errstr,*) "lower ML limit for source detections = ",maxlim
    call message(AppMsg,Always)
    !
    lunin = GetLun() ! in esas_mod in goflib
    open(unit=lunin,form='formatted',status='old',&
         file=trim(adjustl(dirs_F)),iostat=ios)
    if (ios .ne. 0) then
       call error ("badTextFile",&
            "Input directories text file does not exist or is unreadable")
    endif
    allocate(fullpath(2500)) ! for now
    fullpath = ''
    nGood = 0
    i = 0
    !
    ! Find all the emllists, sanity check, make final doable list.
    ! Max number of files set to 2500, with ~40 sources per file,
    ! Max number of sources is 100000. This can be changed ???!!!
    ! If runemlfill is true, look for if done, if not, then do
    !
    do while (ios .eq. 0)
       read(lunin,'(A)',iostat=ios) currdir
       i = i + 1
       if (i .gt. 2500) then ! change this???!!!
          call warning('MaxDirs2500',&
               "Maximum directories is 2500, will only do first 2500")
          nGood = 2500
          exit
       end if
       write(errstr,*) "Current directory : ",trim(adjustl(currdir))
       call message(AppMsg,Noisy)
       currpath_F  = trim(adjustl(currdir)) // trim(adjustl(eml_F))
       currpath_FF = trim(adjustl(currdir)) // trim(adjustl(eml_FF))
       !
       inquire(FILE=trim(currpath_F), EXIST=exists) ! check for emllist.fits
       if (.not. exists) then
          inquire(FILE=trim(currpath_FF), EXIST=exists) ! emllist_filled.fits
          if (.not. exists) then
             ! if neither emllist.fits nor emllist_filled.fits exist...
             write(errstr,*)"Neither emllist or emllist_filled files in ",&
                  trim(adjustl(currdir))," skipping..."
             call warning("noEMLLIST",trim(adjustl(errstr)))
             cycle
          endif
       else
          if (runemlfill) then ! relies on default emllist.fits
             command = "emlfill emlin=" // trim(adjustl(currpath_F)) // &
                  " emlout=" // trim(adjustl(currpath_FF))
             write(errstr,*) "emlfill command = "
             call message(AppMsg,Verbose)
             write(errstr,*) trim(adjustl(command))
             call message(AppMsg,verbose)
             call system(trim(adjustl(command)))
             currpath_F = currpath_FF ! new curr file fullpath
          endif
       endif
       ! sanity checks
       status = 0
       warndie = .TRUE. ! 1 = read
       call check_open_set(trim(adjustl(currpath_F)),inEML_S,1,status,warndie)
       if (status .eq. 1) then
          write(errstr,*)trim(adjustl(currpath_F)),&
               ' cannot open or read, disgarding...'
          call warning("badEMLLIST",trim(adjustl(errstr)))
          cycle
       end if
       if (.not.hasBlock(inEML_S,'SRCLIST')) then
          write(errstr,*)trim(adjustl(currpath_F)),&
               ' missing SRCLIST block, disgarding...'
          call warning("badEMLLIST",trim(adjustl(errstr)))
          cycle
       else
          nGood = nGood + 1
          fullpath(i) = trim(adjustl(currpath_F)) ! accumulated 'good ones'
       endif
       call release(inEML_S)
    enddo
    ! free up some wasted memory
    allocate(goodemllist(j))
    do i = 1, nGood
       goodemllist(i) = fullpath(i)
    end do
    deallocate(fullpath)
    !
    if (nGood .lt. 1) then
       call error("noEMLfound",&
            "None of srclist files in dirlist exist or are readable")
    endif
    !
    ! allocate temp arrays, will shrink later, see bottom code
    !
    init = nGood * 200 ! change this???!!!
    allocate(outRA(init),outDEC(init),outRADEC(init),outML(init))
    allocate(outDIST(init),outFLUX(init),outFLUXERR(init),&
         outFLUXS(init),outFLUXSERR(init),outFLUXH(init),outFLUXHERR(init))
    !
    ! loop through all actual good emllist.fits files
    !
    csrc = 0 ! running source number
    do i = 1, nGood
       inEML_F = goodemllist(i)
       status = 0
       warndie = .TRUE. ! 1 = read
       call check_open_set(trim(adjustl(inEML_F)),inEML_S,1,status,warndie)
       inEML_T = table(inEML_S,'SRCLIST', MODIFY)
       nRows = numberOfRows(inEML_T) ! channels or events, depending
       if (mod(nRows,4) .ne. 0) then ! shouldn't happen, but...
          call warning("not4Bands",&
               "Number of rows in emllist not evenly divisible by 4")
       endif
       write(errstr,*)'Current emllist.fits file has ',nRows,' rows.'
       !
       ! extract relevant columns from current emllist
       !
       if (.not.hasColumn(inEML_T,'ML_ID_SRC')) then
          call error("noMLIDSRC","Current EML file does not contain an ML_ID_SRC column.")
       else
          mlidsrc_C = column(inEML_T,'ML_ID_SRC', READ)
          mlidsrc_JP => int32Data(mlidsrc_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'ID_INST')) then
          call error("noIDINST","Current EML file does not contain an ID_INST column.")
       else
          idinst_C = column(inEML_T,'ID_INST', READ)
          idinst_JP => int32Data(idinst_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'ID_BAND')) then
          call error("noIDBAND","Current EML file does not contain an ID_BAND column.")
       else
          idband_C = column(inEML_T,'ID_BAND', READ)
          idband_JP => int32Data(idband_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'SCTS')) then
          call error("noSCTS","Current EML file does not contain a SCTS column.")
       else
          scts_C = column(inEML_T,'SCTS', READ)
          scts_RP => real32Data(scts_C) 
       endif
       call fixnans(scts_RP,nRows) ! converts nulls to zeros
       !
       if (.not.hasColumn(inEML_T,'DET_ML')) then
          call error("noDETML","Current EML file does not contain a DET_ML column.")
       else
          detml_C = column(inEML_T,'DET_ML', READ)
          detml_RP => real32Data(detml_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'FLUX')) then
          call error("noFLUX","Current EML file does not contain a FLUX column.")
       else
          flux_C = column(inEML_T,'FLUX', READ)
          flux_RP => real32Data(flux_C) 
       endif
       call fixnans(flux_RP,nRows) ! converts nulls to zeros
       !
       if (.not.hasColumn(inEML_T,'FLUX_ERR')) then
          call error("noFLUXERR","Current file does not contain a FLUX_ERR column.")
       else
          fluxerr_C = column(inEML_T,'FLUX_ERR', READ)
          fluxerr_RP => real32Data(fluxerr_C)        
       endif
       call fixnans(fluxerr_RP,nRows) ! converts nulls to zeros
       !
       if (.not.hasColumn(inEML_T,'RATE')) then
          call error("noRATE","Current EML file does not contain a RATE column.")
       else
          rate_C = column(inEML_T,'RATE', READ)
          rate_RP => real32Data(rate_C) 
       endif
       call fixnans(rate_RP,nRows) ! converts nulls to zeros
       !
       if (.not.hasColumn(inEML_T,'RATE_ERR')) then
          call error("noRATEERR","Current file does not contain a RATE_ERR column.")
       else
          rateerr_C = column(inEML_T,'RATE_ERR', READ)
          rateerr_RP => real32Data(rateerr_C)        
       endif
       call fixnans(rateerr_RP,nRows) ! converts nulls to zeros
       !
       if (.not.hasColumn(inEML_T,'RA')) then
          call error("noRA","Current EML file does not contain an RA column.")
       else
          ra_C = column(inEML_T,'RA', READ)
          ra_DP => real64Data(ra_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'DEC')) then
          call error("noDEC","Current EML file does not contain a DEC column.")
       else
          dec_C = column(inEML_T,'DEC', READ)
          dec_DP => real64Data(dec_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'RADEC_ERR')) then
          call error("noRADECERR",&
               "Current EML file does not contain an RADEC_ERR column.")
       else
          radec_C = column(inEML_T,'RADEC_ERR', READ)
          radec_RP => real32Data(radec_C) 
       endif
       !
       if (.not.hasColumn(inEML_T,'DIST_NN')) then
          call error("noDISTNN","Current EML file does not contain a DIST_NN column.")
       else
          dist_C = column(inEML_T,'DIST_NN', READ)
          dist_RP => real32Data(dist_C) 
       endif
       tr1  = 0. 
       tf1  = 0. 
       tre1 = 0. 
       tfe1 = 0. 
       tnr1 = 0. 
       tnf1 = 0.
       tr2  = 0.
       tf2  = 0.
       tre2 = 0.
       tfe2 = 0.
       tnr2 = 0.
       tnf2 = 0.
       !  Loop through the source entries
       crow = 0 ! current row of current file
       do ! every row in current emllist file
          if (crow .gt. nRows) exit ! the current emllist
          if (idband_JP(crow) .eq. 0) then
             outRA(csrc)      = ra_DP(crow)
             outDEC(csrc)     = dec_DP(crow)
             outRADEC(csrc)   = radec_RP(crow)
             outML(csrc)      = detml_RP(crow)
             outDIST(csrc)    = dist_RP(crow)
             if (outML(csrc) .lt. maxlikelim) then
                flux_RP(crow) = 0.
                fluxerr_RP(crow) = 0.
             end if ! not true for fluxs and fluxh???!!!
             outFLUX(csrc)    = flux_RP(crow)
             outFLUXERR(csrc) = fluxerr_RP(crow)
          else
             do j = crow+1,crow+3
                if (idband_JP(crow) .eq. 1) then
                   tf1  = tf1  + scts_RP(j)*flux_RP(j)
                   tfe1 = tfe1 + (scts_RP(j)*fluxerr_RP(j))**2
                   tnf1 = tnf1 + scts_RP(j)
                elseif (idband_JP(crow) .eq. 2) then
                   tf2  = tf2  + scts_RP(j)*flux_RP(j)
                   tfe2 = tfe2 + (scts_RP(j)*fluxerr_RP(j))**2
                   tnf2 = tnf2 + scts_RP(j)
                end if
             end do
          end if
          !
          if(tnf1 .gt. 0.0) tf1 = tf1/tnf1
          if(tnf1 .gt. 0.0) tfe1 = sqrt(tfe1)/tnf1
          if(tnf2 .gt. 0.0) tf2 = tf2/tnf2
          if(tnf2 .gt. 0.0) tfe2 = sqrt(tfe2)/tnf2
          !
          outFLUXS(csrc) = tf1
          outFLUXSERR(csrc) = tfe1
          outFLUXH(csrc) = tf2
          outFLUXHERR(csrc) = tfe2
          !
          crow = crow + 4
          csrc = csrc + 1
       end do ! loop thru single file's sources 
       
    end do ! loop thru all 'good' emllists
    
    !  Sort through the sources, compare the all bands, all instruments
    !  dectection information
    
    do i=1,csrc-1
       if(outFLUX(i) .gt. 0.0) then
          dlong1 = 0.017453293*outRA(i)
          dlat1 = 0.017453293*outDEC(i)
          do ii=i+1,csrc
             if(outFLUX(ii) .gt. 0.0) then
                
                !  Get the angle between the two sources

                dlong2 = 0.017453293*outRA(ii)
                dlat2 = 0.017453293*outDEC(ii)
                call angsep (dlong1,dlat1,dlong2,dlat2,dang)
                
                ! Calculate the position uncertainty, add in quadrature
                ! the statistical uncertainty from the source list and
                ! the astrometric uncertainty (2") for each source,
                ! and then add them together in quadrature.

                punc = sqrt(4.0 + outRADEC(i)*outRADEC(i) &
                     + 4.0 + outRADEC(ii)*outRADEC(ii))
                punc = punc*4.84814e-6

                k = 1.5*dang/(punc) + 1
                if(k .eq. 0) k = 1
                if(k .gt. 1000) k = 1000
                hist1(k) = hist1(k) + 1

                ! If the angle is less than 1.5x the uncertainty
                ! in the distance between the two detections get
                ! rid of the fainter detection

                if(dang .lt. 1.5*punc) then
                   if(outFLUX(i) .lt. outFLUX(ii)) then
                      outFLUX(i) = 0.
                   else
                      outFLUX(ii) = 0.
                   endif
                else
                   if(dang .lt. dist(i)) dist(i) = dang
                   if(dang .lt. dist(ii)) dist(ii) = dang
                endif
             endif
          enddo
       endif
    enddo
         
    !  Now get rid of the fainter detections from the list
         
    ii = 0
    do i=1,csrc
       if(flux(i) .gt. 0.0) then
          ii = ii + 1
          ra(ii) = ra(i)
          dec(ii) = dec(i)
          radec(ii) = radec(i)
          ml(ii) = ml(i)
          k = dist(i)/(2.0*2.4240684d-5) + 1
          if(k .eq. 0) k = 1
          if(k .gt. 1000) k = 1000
          hist2(k) = hist2(k) + 1
          dist(ii) = 2.06265e5*dist(i)
          outFLUX(ii)     = outFLUX(i)
          outFLUXERR(ii)  = outFLUXERR(i)
          outFLUXS(ii)    = outFLUXS(i)
          outFLUXSERR(ii) = outFLUXSERR(i)
          outFLUXH(ii)    = outFLUXH(i)
          outFLUXHERR(ii) = outFLUXHERR(i)
       endif
    enddo
    nGood = ii
    print *, 'Total Unique Sources: ', nGood
    !
    ! open the output merged source list
    !
    write(errstr,*) 'Found ',nGood,' mergeable srclist files to process.'
    call message(AppMsg,Always)
    write(errstr,*) 'Found ',csrc,' total sources in all srclists.'
    call message(AppMsg,Always)
    status = 0
    warndie = .TRUE. ! 3 = create
    call check_open_set(trim(adjustl(outEML_F)),outEML_S,3,status,warndie)
    outEML_T = addTable(outEML_S,"SRCLIST",nGood)
    !
    ! add the columns
    !
    outRA_C = addColumn(outEML_T,"RA",REAL64,units="deg",&
         comment="Source Right Ascension")
    outRA_DP => real64Data(outRA_C)
    outDEC_C = addColumn(outEML_T,"DEC",REAL64,units="deg",&
         comment="Source Declination")
    outDEC_DP => real64Data(outDEC_C)
    outRADEC_C = addColumn(outEML_T,"RADEC_ER",REAL32,units="arcsec",&
         comment="Source Position Error")
    outRADEC_RP => real32Data(outRADEC_C)
    outML_C = addColumn(outEML_T,"DET_ML",REAL32,units="cgs",&
         comment="Maximum Likelihood Detection Value")
    outML_RP => real32Data(outML_C)
    !
    outFLUX_C = addColumn(outEML_T,"FLUX_TOT",REAL32,units="cgs",&
         comment="Source Total Flux")
    outFLUX_RP => real32Data(outFLUX_C)
    outFLUXERR_C = addColumn(outEML_T,"FL_T_ERR",REAL32,units="cgs",&
         comment="Source Total Flux Uncertainty")
    outFLUXERR_RP => real32Data(outFLUXERR_C)
    !
    outFLUXS_C = addColumn(outEML_T,"FLUX_S",REAL32,units="cgs",&
         comment="Source Soft Flux")
    outFLUXS_RP => real32Data(outFLUXS_C)
    outFLUXSERR_C = addColumn(outEML_T,"FL_S_ERR",REAL32,units="cgs",&
         comment="Source Soft Flux Uncertainty")
    outFLUXSERR_RP => real32Data(outFLUXSERR_C)
    !
    outFLUXH_C = addColumn(outEML_T,"FLUX_H",REAL32,units="cgs",&
         comment="Source Hard Flux")
    outFLUXH_RP => real32Data(outFLUXH_C)
    outFLUXH_C = addColumn(outEML_T,"FLUX_H_ERR",REAL32,units="cgs",&
         comment="Source Hard Flux Uncertainty")
    outFLUXHERR_RP => real32Data(outFLUXHERR_C)
    !
    outDIST_C = addColumn(outEML_T,"DIST_NN",REAL32,units="arcsec",&
         comment="Separation Angle from the Nearest Source")
    outDIST_RP => real32Data(outDIST_C)
    !
    ! Fill the out pointers
    !
    do i = 0, nGood-1
       outRA_DP(i) = outRA(i+1)
       outDEC_DP(i) = outDEC(i+1)
       outRADEC_RP(i) = outRADEC(i+1)
       outML_RP(i) = outML(i+1)
       outFLUX_RP(i) = outFLUX(i+1)
       outFLUXERR_RP(i) = outFLUXERR(i+1)
       outFLUXS_RP(i) = outFLUXS(i+1)
       outFLUXSERR_RP(i) = outFLUXSERR(i+1)
       outFLUXH_RP(i) = outFLUXH(i+1)
       outFLUXHERR_RP(i) = outFLUXHERR(i+1)
       outDIST_RP(i) = outDIST(i+1)
    end do
    !
    call release(outEML_S)
    !
    ! Done with merged fits source list, now do diagnostic QDP plots
    ! QDP1
    allocate (hist1(1000),hist2(1000)) ! fix this ???!!!
    warndie = .FALSE.
    doit = .FALSE.
    outQDP_F = "sourcesep1.qdp"
    call inquirer(outQDP_F,warndie,doit) ! check clobber, exist, die if necessary
    if (doit) then 
       lunout = GetLun()
       open(unit=lunout,status='new',form='formatted',&
            file=trim(adjustl(outQDP_F)),iostat=ios)
       if (ios .ne. 0) then
          call error("badQDP1",&
               "QDP output file could not be opened for write")
       else
          histmax = 0.0
          histrange = 0.0
          do i=2,999 ! fix this???!!!
             if(hist1(i) .gt. histmax) histmax = hist1(i)
             if(hist1(i) .gt. 0) histrange = i
          enddo
          histmax = 1.05 * histmax
          histrange = histrange * 10.0 / 60.0
          write(lunout,*) 'viewport 0.2 0.11 0.95 0.9'
          write(lunout,*) 'font roman'
          write(lunout,*) 'cs 1.5'
          write(lunout,*) 'la t Source Separation'
          write(lunout,*) 'la x Separation (units of position error)'
          write(lunout,*) 'la y Number'
          write(lunout,*) 'la f'
          write(lunout,*) 'col 1 on 2'
          write(lunout,*) 'line step'
          write(lunout,*) 'r 0 ', histrange, '  0 ', histmax
          do i=1,1000
             write(lunout,*) (i-0.5)*10.0/60.0, hist1(i)
          enddo
          close(lunout)
       endif
    else
       call warning("noQDP1","Skipping sourcesep1.qdp creation")
    end if
    !
    !
    ! QDP2
    warndie = .FALSE.
    doit = .FALSE.
    outQDP_F = "sourcesep2.qdp"
    call inquirer(outQDP_F,warndie,doit) ! check clobber, exist, die if necessary
    if (doit) then 
       lunout = GetLun()
       open(unit=lunout,status='new',form='formatted',&
            file=trim(adjustl(outQDP_F)),iostat=ios)
       if (ios .ne. 0) then
          call error("badQDP2",&
               "QDP output file could not be opened for write")
       else
          histmax = 0.0
          histrange = 0.0
          do i=2,999
             if(hist2(i) .gt. histmax) histmax = hist2(i)
             if(hist2(i) .gt. 0) histrange = i
          enddo
          histmax = 1.05*histmax
          histrange = histrange*10.0/60.0
          
          open(lunout,status='unknown',form='formatted', &
               file='sep2.qdp')
          write(lunout,*) 'viewport 0.2 0.11 0.95 0.9'
          write(lunout,*) 'font roman'
          write(lunout,*) 'cs 1.5'
          write(lunout,*) 'la t Minimum Source Separation'
          write(lunout,*) 'la x Separation (arc minutes)'
          write(lunout,*) 'la y Number'
          write(lunout,*) 'la f'
          write(lunout,*) 'col 1 on 2'
          write(lunout,*) 'line step'
          write(lunout,*) 'r 0 ', histrange, '  0 ', histmax
          do i=1,1000
             write(lunout,*) (i-0.5)*10.0/60.0, hist2(i)
          enddo
          close(lunout)
       endif
    else
       call warning("noQDP2","Skipping sourcesep2.qdp creation")
    end if
    !
    !
    ! QDP3
    warndie = .FALSE.
    doit = .FALSE.
    outQDP_F = "positions.qdp"
    call inquirer(outQDP_F,warndie,doit) ! check clobber, exist, die if necessary
    if (doit) then 
       lunout = GetLun()
       open(unit=lunout,status='new',form='formatted',&
            file=trim(adjustl(outQDP_F)),iostat=ios)
       if (ios .ne. 0) then
          call error("badQDP3",&
               "QDP output file could not be opened for write")
       else
          ramin = 1000.0
          ramax = 0.0
          decmin = 90.0
          decmax = -90.0
          do i=1,nrows
             if(ra(i) .lt. ramin) ramin = ra(i)
             if(ra(i) .gt. ramax) ramax = ra(i)
             if(dec(i) .lt. decmin) decmin = dec(i)
             if(dec(i) .gt. decmax) decmax = dec(i)
          enddo
          if((ramax - ramin) .gt. 180.0) then
             do i=1,nrows
                if(ra(i) .gt. 180.0) ra(i) = ra(i) - 360.0
             enddo
             ramin = 1000.0
             ramax = -1000.0
             do i=1,nrows
                if(ra(i) .lt. ramin) ramin = ra(i)
                if(ra(i) .gt. ramax) ramax = ra(i)
             enddo
          endif
          delra = ramax - ramin
          deldec = decmax - decmin
          raave = (ramax + ramin)/2.0
          decave = (decmax + decmin)/2.0
          scale = cos(decave*3.14159265/180.0)
          rarange = scale*delra
          if(rarange .lt. deldec) then
             ramin = raave - 1.025*0.5*delra/scale
             ramax = raave + 1.025*0.5*delra/scale
             decmin = decave - 1.05*deldec/2.0
             decmax = decave + 1.05*deldec/2.0
          else
             ramin = raave - 1.05*delra/2.0
             ramax = raave + 1.05*delra/2.0
             decmin = decave - 0.5*delra*rarange/deldec
             decmax = decave + 0.5*delra*rarange/deldec
          endif
          !
          write(lunout,*) 'viewport 0.2 0.11 0.78515 0.9'
          write(lunout,*) 'font roman'
          write(lunout,*) 'cs 1.5'
          write(lunout,*) 'la t Source Positions'
          write(lunout,*) 'la x Right Ascension'
          write(lunout,*) 'la y Declination'
          write(lunout,*) 'la f'
          write(lunout,*) 'col 1 on 2'
          write(lunout,*) 'line off 2'
          write(lunout,*) 'mark 2 on 2'
          write(lunout,*) 'r ', ramax, ramin, decmin, decmax
          do i=1,nrows
             write(lunout,*) ra(i), dec(i)
          enddo
          close(lunout)
       endif
    else
       call warning("noQDP2","Skipping positions.qdp creation")
    end if

    return

  end subroutine emlmerge

  subroutine fixnans(inout_RP,nRows)
    !
    ! fixes all ieee nulls in relevant emllist columns
    !
    real, intent(inout), dimension(:), pointer :: inout_RP
    integer :: nRows, i
    !
    do i=0,nRows-1
       if (ieee_is_nan(inout_RP(i))) then ! is the corresponding value NULL?
          inout_RP(i) = 0.
       end if
    end do
    return
  end subroutine fixnans

  subroutine inquirer(file,warndie,doit)
    !
    character(len=*) :: file
    logical :: exists, warndie, doit
    !
    inquire(FILE=trim(adjustl(file)), EXIST=exists)
    if (exists) then
       if (clobber()) then
          write(errstr,*)"Output file ",trim(adjustl(file)),&
               " exists but clobber set, overwriting."
          call message(AppMsg,Always)
          call deleteDataSet(trim(adjustl(file)))
       else
          if (warndie) then
             errstr = "Critical I/O file " // trim(adjustl(file)) // &
                  " exists but noclobber set. Exiting"
             call error("noClobber",trim(adjustl(errstr)))
          else
             errstr = "NON-critical I/O file " // trim(adjustl(file)) // &
                  " exists but noclobber set. skiping"
             call warning("noClobber",trim(adjustl(errstr)))
             doit = .FALSE.
          end if
       endif
    endif
    
  end subroutine inquirer
  
end module emlmerge_mod

! for later
! allocate(tmp_arr(2*size(array)))
! tmp_arr(1:size(array)) = array
! deallocate(array)
! move_alloc(tmp_arr, array)
! tmp_arr is now deallocated
